@@ -5,6 +5,13 @@
 use std::net::IpAddr;
 use std::str::FromStr;
 
+use asn1::IA5String;
+
+// RFC 2822 3.2.4
+static ATEXT_CHARS: &str = "!#$%&'*+-/=?^_`{|}~";
+
+/// Represents a DNS name can be used in X.509 name matching.
+///
 /// A `DNSName` is an `asn1::IA5String` with additional invariant preservations
 /// per [RFC 5280 4.2.1.6], which in turn uses the preferred name syntax defined
 /// in [RFC 1034 3.5] and amended in [RFC 1123 2.1].
@@ -75,6 +82,17 @@ impl<'a> DNSName<'a> {
     fn rlabels(&self) -> impl Iterator<Item = &'_ str> {
         self.as_str().rsplit('.')
     }
+
+    /// Returns true if this domain is a subdomain of the other domain.
+    fn is_subdomain_of(&self, other: &DNSName<'_>) -> bool {
+        // NOTE: This is nearly identical to `DNSConstraint::matches`,
+        // except that the subdomain must be strictly longer than the parent domain.
+        self.as_str().len() > other.as_str().len()
+            && self
+                .rlabels()
+                .zip(other.rlabels())
+                .all(|(a, o)| a.eq_ignore_ascii_case(o))
+    }
 }
 
 impl PartialEq for DNSName<'_> {
@@ -84,6 +102,9 @@ impl PartialEq for DNSName<'_> {
     }
 }
 
+/// Represents either a DNS name or a DNS wildcard for use in X.509 name
+/// matching.
+///
 /// A `DNSPattern` represents a subset of the domain name wildcard matching
 /// behavior defined in [RFC 6125 6.4.3]. In particular, all DNS patterns
 /// must either be exact matches (post-normalization) *or* a single wildcard
@@ -298,9 +319,89 @@ impl IPConstraint {
     }
 }
 
+/// An `RFC822Name` represents an email address, as defined in [RFC 822 6.1]
+/// and as amended by [RFC 2821 4.1.2]. In particular, it represents the `Mailbox`
+/// rule from RFC 2821's grammar.
+///
+/// This type does not currently support the quoted local-part form; email
+/// addresses that use this form will be rejected.
+///
+/// [RFC 822 6.1]: https://datatracker.ietf.org/doc/html/rfc822#section-6.1
+/// [RFC 2821 4.1.2]: https://datatracker.ietf.org/doc/html/rfc2821#section-4.1.2
+#[derive(PartialEq)]
+pub struct RFC822Name<'a> {
+    pub mailbox: IA5String<'a>,
+    pub domain: DNSName<'a>,
+}
+
+impl<'a> RFC822Name<'a> {
+    pub fn new(value: &'a str) -> Option<Self> {
+        // Mailbox = Local-part "@" Domain
+        // Both must be present.
+        let (local_part, domain) = value.split_once('@')?;
+        let local_part = IA5String::new(local_part)?;
+
+        // Local-part = Dot-string / Quoted-string
+        // NOTE(ww): We do not support the Quoted-string form, for now.
+        //
+        // Dot-string: Atom *("." Atom)
+        // Atom = 1*atext
+        //
+        // NOTE(ww): `atext`'s production is in RFC 2822 3.2.4.
+        for component in local_part.as_str().split('.') {
+            if component.is_empty()
+                || !component
+                    .chars()
+                    .all(|c| c.is_ascii_alphanumeric() || ATEXT_CHARS.contains(c))
+            {
+                return None;
+            }
+        }
+
+        Some(Self {
+            mailbox: local_part,
+            domain: DNSName::new(domain)?,
+        })
+    }
+}
+
+/// An `RFC822Constraint` represents a Name Constraint on email addresses.
+pub enum RFC822Constraint<'a> {
+    /// A constraint for an exact match on a specific email address.
+    Exact(RFC822Name<'a>),
+    /// A constraint for any mailbox on a particular domain.
+    OnDomain(DNSName<'a>),
+    /// A constraint for any mailbox *within* a particular domain.
+    /// For example, `InDomain("example.com")` will match `foo@bar.example.com`
+    /// but not `foo@example.com`, since `bar.example.com` is in `example.com`
+    /// but `example.com` is not within itself.
+    InDomain(DNSName<'a>),
+}
+
+impl<'a> RFC822Constraint<'a> {
+    pub fn new(constraint: &'a str) -> Option<Self> {
+        if let Some(constraint) = constraint.strip_prefix('.') {
+            Some(Self::InDomain(DNSName::new(constraint)?))
+        } else if let Some(email) = RFC822Name::new(constraint) {
+            Some(Self::Exact(email))
+        } else {
+            Some(Self::OnDomain(DNSName::new(constraint)?))
+        }
+    }
+
+    pub fn matches(&self, email: &RFC822Name<'_>) -> bool {
+        match self {
+            Self::Exact(pat) => pat == email,
+            Self::OnDomain(pat) => &email.domain == pat,
+            Self::InDomain(pat) => email.domain.is_subdomain_of(pat),
+        }
+    }
+}
+
 #[cfg(test)]
 mod tests {
-    use crate::types::{DNSConstraint, DNSName, DNSPattern, IPAddress, IPConstraint};
+    use super::RFC822Constraint;
+    use crate::types::{DNSConstraint, DNSName, DNSPattern, IPAddress, IPConstraint, RFC822Name};
 
     #[test]
     fn test_dnsname_debug_trait() {
@@ -392,6 +493,33 @@ mod tests {
         );
     }
 
+    #[test]
+    fn test_dnsname_is_subdomain_of() {
+        for (sup, sub, check) in &[
+            // good cases
+            ("example.com", "sub.example.com", true),
+            ("example.com", "a.b.example.com", true),
+            ("sub.example.com", "sub.sub.example.com", true),
+            ("sub.example.com", "sub.sub.sub.example.com", true),
+            ("com", "example.com", true),
+            ("example.com", "com.example.com", true),
+            ("example.com", "com.example.example.com", true),
+            // bad cases
+            ("example.com", "example.com", false),
+            ("example.com", "com", false),
+            ("sub.example.com", "example.com", false),
+            ("sub.sub.example.com", "sub.sub.example.com", false),
+            ("sub.sub.example.com", "example.com", false),
+            ("com.example.com", "com.example.com", false),
+            ("com.example.example.com", "com.example.example.com", false),
+        ] {
+            let sup = DNSName::new(sup).unwrap();
+            let sub = DNSName::new(sub).unwrap();
+
+            assert_eq!(sub.is_subdomain_of(&sup), *check);
+        }
+    }
+
     #[test]
     fn test_dnspattern_new() {
         assert_eq!(DNSPattern::new("*"), None);
@@ -587,4 +715,153 @@ mod tests {
         assert!(!ipv6_128.matches(&IPAddress::from_str("2600::ff00:dede").unwrap()));
         assert!(!ipv6_128.matches(&IPAddress::from_str("2600:db8::ff00:0").unwrap()));
     }
+
+    #[test]
+    fn test_rfc822name() {
+        for bad_case in &[
+            "",
+            // Missing local-part.
+            "@example.com",
+            " @example.com",
+            "  @example.com",
+            // Missing domain cases.
+            "foo",
+            "foo@",
+            "foo@ ",
+            "foo@  ",
+            // Invalid domains.
+            "foo@!!!",
+            "foo@white space",
+            "foo@🙈",
+            // Invalid local part (empty mailbox sections).
+            ".@example.com",
+            "foo.@example.com",
+            ".foo@example.com",
+            ".foo.@example.com",
+            ".f.o.o.@example.com",
+            // Invalid local part (@ in mailbox).
+            "lol@lol@example.com",
+            "lol\\@lol@example.com",
+            "example@example.com@example.com",
+            "@@example.com",
+            // Invalid local part (invalid characters).
+            "lol\"lol@example.com",
+            "lol;lol@example.com",
+            "🙈@example.com",
+            // Intentionally unsupported quoted local parts.
+            "\"validbutunsupported\"@example.com",
+        ] {
+            assert!(RFC822Name::new(bad_case).is_none());
+        }
+
+        // Each good case is (address, (mailbox, domain)).
+        for (address, (mailbox, domain)) in &[
+            // Normal mailboxes.
+            ("foo@example.com", ("foo", "example.com")),
+            ("foo.bar@example.com", ("foo.bar", "example.com")),
+            ("foo.bar.baz@example.com", ("foo.bar.baz", "example.com")),
+            ("1.2.3.4.5@example.com", ("1.2.3.4.5", "example.com")),
+            // Mailboxes with special but valid characters.
+            ("{legal}@example.com", ("{legal}", "example.com")),
+            ("{&*.legal}@example.com", ("{&*.legal}", "example.com")),
+            ("``````````@example.com", ("``````````", "example.com")),
+            ("hello?@sub.example.com", ("hello?", "sub.example.com")),
+        ] {
+            let parsed = RFC822Name::new(&address).unwrap();
+            assert_eq!(&parsed.mailbox.as_str(), mailbox);
+            assert_eq!(&parsed.domain.as_str(), domain);
+        }
+    }
+
+    #[test]
+    fn test_rfc822constraint_new() {
+        for (case, valid) in &[
+            // good cases
+            ("foo@example.com", true),
+            ("foo.bar@example.com", true),
+            ("foo!bar@example.com", true),
+            ("example.com", true),
+            ("sub.example.com", true),
+            ("foo@sub.example.com", true),
+            ("foo.bar@sub.example.com", true),
+            ("foo!bar@sub.example.com", true),
+            (".example.com", true),
+            (".sub.example.com", true),
+            // bad cases
+            ("@example.com", false),
+            ("@@example.com", false),
+            ("foo@.example.com", false),
+            (".foo@example.com", false),
+            (".foo.@example.com", false),
+            ("foo.@example.com", false),
+            ("invaliddomain!", false),
+            ("..example.com", false),
+            ("foo..example.com", false),
+            (".foo..example.com", false),
+            ("..foo..example.com", false),
+        ] {
+            assert_eq!(RFC822Constraint::new(case).is_some(), *valid);
+        }
+    }
+
+    #[test]
+    fn test_rfc822constraint_matches() {
+        {
+            let exact = RFC822Constraint::new("foo@example.com").unwrap();
+
+            // Ordinary exact match.
+            assert!(exact.matches(&RFC822Name::new("foo@example.com").unwrap()));
+            // Case changes are okay in the domain.
+            assert!(exact.matches(&RFC822Name::new("foo@EXAMPLE.com").unwrap()));
+
+            // Case changes are not okay in the mailbox.
+            assert!(!exact.matches(&RFC822Name::new("Foo@example.com").unwrap()));
+            assert!(!exact.matches(&RFC822Name::new("FOO@example.com").unwrap()));
+
+            // Different mailboxes and domains do not match.
+            assert!(!exact.matches(&RFC822Name::new("foo.bar@example.com").unwrap()));
+            assert!(!exact.matches(&RFC822Name::new("foo@sub.example.com").unwrap()));
+        }
+
+        {
+            let on_domain = RFC822Constraint::new("example.com").unwrap();
+
+            // Ordinary domain matches.
+            assert!(on_domain.matches(&RFC822Name::new("foo@example.com").unwrap()));
+            assert!(on_domain.matches(&RFC822Name::new("bar@example.com").unwrap()));
+            assert!(on_domain.matches(&RFC822Name::new("foo.bar@example.com").unwrap()));
+            assert!(on_domain.matches(&RFC822Name::new("foo!bar@example.com").unwrap()));
+            // Case changes are okay in the domain and in the mailbox,
+            // since any mailbox on the domain is okay.
+            assert!(on_domain.matches(&RFC822Name::new("foo@EXAMPLE.com").unwrap()));
+            assert!(on_domain.matches(&RFC822Name::new("FOO@example.com").unwrap()));
+
+            // Subdomains and other domains do not match.
+            assert!(!on_domain.matches(&RFC822Name::new("foo@sub.example.com").unwrap()));
+            assert!(!on_domain.matches(&RFC822Name::new("foo@localhost").unwrap()));
+        }
+
+        {
+            let in_domain = RFC822Constraint::new(".example.com").unwrap();
+
+            // Any subdomain and mailbox matches.
+            assert!(in_domain.matches(&RFC822Name::new("foo@sub.example.com").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("foo@sub.sub.example.com").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("foo@com.example.example.com").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("foo.bar@com.example.example.com").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("foo!bar@com.example.example.com").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("bar@com.example.example.com").unwrap()));
+            // Case changes are okay in the subdomains and in the mailbox, since any mailbox
+            // in the domain is okay.
+            assert!(in_domain.matches(&RFC822Name::new("foo@SUB.example.com").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("foo@sub.EXAMPLE.com").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("foo@sub.example.COM").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("FOO@sub.example.COM").unwrap()));
+            assert!(in_domain.matches(&RFC822Name::new("FOO@sub.example.com").unwrap()));
+
+            // Superdomains and other domains do not match.
+            assert!(!in_domain.matches(&RFC822Name::new("foo@example.com").unwrap()));
+            assert!(!in_domain.matches(&RFC822Name::new("foo@com").unwrap()));
+        }
+    }
 }
