@@ -5,21 +5,22 @@
 use std::collections::hash_map::DefaultHasher;
 use std::hash::{Hash, Hasher};
 
-use pyo3::ToPyObject;
+use pyo3::types::{PyAnyMethods, PyDictMethods};
 
 use crate::backend::utils;
 use crate::buf::CffiBuf;
 use crate::error::{CryptographyError, CryptographyResult};
+use crate::x509::common::cstr_from_literal;
 use crate::{exceptions, types};
 
-#[pyo3::prelude::pyclass(frozen, module = "cryptography.hazmat.bindings._rust.openssl.ec")]
+#[pyo3::pyclass(frozen, module = "cryptography.hazmat.bindings._rust.openssl.ec")]
 pub(crate) struct ECPrivateKey {
     pkey: openssl::pkey::PKey<openssl::pkey::Private>,
     #[pyo3(get)]
     curve: pyo3::Py<pyo3::PyAny>,
 }
 
-#[pyo3::prelude::pyclass(frozen, module = "cryptography.hazmat.bindings._rust.openssl.ec")]
+#[pyo3::pyclass(frozen, module = "cryptography.hazmat.bindings._rust.openssl.ec")]
 pub(crate) struct ECPublicKey {
     pkey: openssl::pkey::PKey<openssl::pkey::Public>,
     #[pyo3(get)]
@@ -28,14 +29,14 @@ pub(crate) struct ECPublicKey {
 
 fn curve_from_py_curve(
     py: pyo3::Python<'_>,
-    py_curve: &pyo3::PyAny,
+    py_curve: pyo3::Bound<'_, pyo3::PyAny>,
     allow_curve_class: bool,
 ) -> CryptographyResult<openssl::ec::EcGroup> {
-    if !py_curve.is_instance(types::ELLIPTIC_CURVE.get(py)?)? {
+    if !py_curve.is_instance(&types::ELLIPTIC_CURVE.get(py)?)? {
         if allow_curve_class {
             let warning_cls = types::DEPRECATED_IN_42.get(py)?;
-            let warning_msg = "Curve argument must be an instance of an EllipticCurve class. Did you pass a class by mistake? This will be an exception in a future version of cryptography.";
-            pyo3::PyErr::warn(py, warning_cls, warning_msg, 1)?;
+            let message = cstr_from_literal!("Curve argument must be an instance of an EllipticCurve class. Did you pass a class by mistake? This will be an exception in a future version of cryptography");
+            pyo3::PyErr::warn(py, &warning_cls, message, 1)?;
         } else {
             return Err(CryptographyError::from(
                 pyo3::exceptions::PyTypeError::new_err("curve must be an EllipticCurve instance"),
@@ -43,8 +44,8 @@ fn curve_from_py_curve(
         }
     }
 
-    let curve_name = py_curve.getattr(pyo3::intern!(py, "name"))?.extract()?;
-    let nid = match curve_name {
+    let py_curve_name = py_curve.getattr(pyo3::intern!(py, "name"))?;
+    let nid = match &*py_curve_name.extract::<pyo3::pybacked::PyBackedStr>()? {
         "secp192r1" => openssl::nid::Nid::X9_62_PRIME192V1,
         "secp224r1" => openssl::nid::Nid::SECP224R1,
         "secp256r1" => openssl::nid::Nid::X9_62_PRIME256V1,
@@ -73,7 +74,7 @@ fn curve_from_py_curve(
         #[cfg(not(CRYPTOGRAPHY_IS_BORINGSSL))]
         "brainpoolP512r1" => openssl::nid::Nid::BRAINPOOL_P512R1,
 
-        _ => {
+        curve_name => {
             return Err(CryptographyError::from(
                 exceptions::UnsupportedAlgorithm::new_err((
                     format!("Curve {curve_name} is not supported"),
@@ -89,16 +90,7 @@ fn curve_from_py_curve(
 fn py_curve_from_curve<'p>(
     py: pyo3::Python<'p>,
     curve: &openssl::ec::EcGroupRef,
-) -> CryptographyResult<&'p pyo3::PyAny> {
-    let name = curve
-        .curve_name()
-        .ok_or_else(|| {
-            pyo3::exceptions::PyValueError::new_err(
-                "ECDSA keys with explicit parameters are unsupported at this time",
-            )
-        })?
-        .short_name()?;
-
+) -> CryptographyResult<pyo3::Bound<'p, pyo3::PyAny>> {
     if curve.asn1_flag() == openssl::ec::Asn1Flag::EXPLICIT_CURVE {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyValueError::new_err(
@@ -107,9 +99,11 @@ fn py_curve_from_curve<'p>(
         ));
     }
 
+    let name = curve.curve_name().unwrap().short_name()?;
+
     types::CURVE_TYPES
         .get(py)?
-        .extract::<&pyo3::types::PyDict>()?
+        .extract::<pyo3::Bound<'_, pyo3::types::PyDict>>()?
         .get_item(name)?
         .ok_or_else(|| {
             CryptographyError::from(exceptions::UnsupportedAlgorithm::new_err((
@@ -132,8 +126,8 @@ fn check_key_infinity(
     Ok(())
 }
 
-#[pyo3::prelude::pyfunction]
-fn curve_supported(py: pyo3::Python<'_>, py_curve: &pyo3::PyAny) -> bool {
+#[pyo3::pyfunction]
+fn curve_supported(py: pyo3::Python<'_>, py_curve: pyo3::Bound<'_, pyo3::PyAny>) -> bool {
     curve_from_py_curve(py, py_curve, false).is_ok()
 }
 
@@ -161,11 +155,12 @@ pub(crate) fn public_key_from_pkey(
         curve: curve.into(),
     })
 }
-#[pyo3::prelude::pyfunction]
+#[pyo3::pyfunction]
+#[pyo3(signature = (curve, backend=None))]
 fn generate_private_key(
     py: pyo3::Python<'_>,
-    curve: &pyo3::PyAny,
-    backend: Option<&pyo3::PyAny>,
+    curve: pyo3::Bound<'_, pyo3::PyAny>,
+    backend: Option<pyo3::Bound<'_, pyo3::PyAny>>,
 ) -> CryptographyResult<ECPrivateKey> {
     let _ = backend;
 
@@ -178,13 +173,13 @@ fn generate_private_key(
     })
 }
 
-#[pyo3::prelude::pyfunction]
+#[pyo3::pyfunction]
 fn derive_private_key(
     py: pyo3::Python<'_>,
-    py_private_value: &pyo3::types::PyLong,
-    py_curve: &pyo3::PyAny,
+    py_private_value: &pyo3::Bound<'_, pyo3::types::PyInt>,
+    py_curve: pyo3::Bound<'_, pyo3::PyAny>,
 ) -> CryptographyResult<ECPrivateKey> {
-    let curve = curve_from_py_curve(py, py_curve, false)?;
+    let curve = curve_from_py_curve(py, py_curve.clone(), false)?;
     let private_value = utils::py_int_to_bn(py, py_private_value)?;
 
     let mut point = openssl::ec::EcPoint::new(&curve)?;
@@ -192,7 +187,9 @@ fn derive_private_key(
     point.mul_generator(&curve, &private_value, &bn_ctx)?;
     let ec = openssl::ec::EcKey::from_private_components(&curve, &private_value, &point)
         .map_err(|_| pyo3::exceptions::PyValueError::new_err("Invalid EC key"))?;
-    check_key_infinity(&ec)?;
+    ec.check_key().map_err(|_| {
+        pyo3::exceptions::PyValueError::new_err("Invalid EC key (key out of range, infinity, etc.)")
+    })?;
     let pkey = openssl::pkey::PKey::from_ec_key(ec)?;
 
     Ok(ECPrivateKey {
@@ -201,13 +198,13 @@ fn derive_private_key(
     })
 }
 
-#[pyo3::prelude::pyfunction]
+#[pyo3::pyfunction]
 fn from_public_bytes(
     py: pyo3::Python<'_>,
-    py_curve: &pyo3::PyAny,
+    py_curve: pyo3::Bound<'_, pyo3::PyAny>,
     data: &[u8],
 ) -> CryptographyResult<ECPublicKey> {
-    let curve = curve_from_py_curve(py, py_curve, false)?;
+    let curve = curve_from_py_curve(py, py_curve.clone(), false)?;
 
     let mut bn_ctx = openssl::bn::BigNumContext::new()?;
     let point = openssl::ec::EcPoint::from_bytes(&curve, data, &mut bn_ctx)
@@ -221,20 +218,23 @@ fn from_public_bytes(
     })
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl ECPrivateKey {
     #[getter]
-    fn key_size<'p>(&'p self, py: pyo3::Python<'p>) -> pyo3::PyResult<&'p pyo3::PyAny> {
-        self.curve.as_ref(py).getattr(pyo3::intern!(py, "key_size"))
+    fn key_size<'p>(
+        &'p self,
+        py: pyo3::Python<'p>,
+    ) -> pyo3::PyResult<pyo3::Bound<'p, pyo3::PyAny>> {
+        self.curve.bind(py).getattr(pyo3::intern!(py, "key_size"))
     }
 
     fn exchange<'p>(
         &self,
         py: pyo3::Python<'p>,
-        algorithm: &pyo3::PyAny,
+        algorithm: pyo3::Bound<'_, pyo3::PyAny>,
         peer_public_key: &ECPublicKey,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
-        if !algorithm.is_instance(types::ECDH.get(py)?)? {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        if !algorithm.is_instance(&types::ECDH.get(py)?)? {
             return Err(CryptographyError::from(
                 exceptions::UnsupportedAlgorithm::new_err((
                     "Unsupported EC exchange algorithm",
@@ -244,7 +244,7 @@ impl ECPrivateKey {
         }
 
         let mut deriver = openssl::derive::Deriver::new(&self.pkey)?;
-        // If `set_peer_ex` is available, we don't valid the key. This is
+        // If `set_peer_ex` is available, we don't validate the key. This is
         // because we already validated it sufficiently when we created the
         // ECPublicKey object.
         #[cfg(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER)]
@@ -257,7 +257,8 @@ impl ECPrivateKey {
             .set_peer(&peer_public_key.pkey)
             .map_err(|_| pyo3::exceptions::PyValueError::new_err("Error computing shared key."))?;
 
-        Ok(pyo3::types::PyBytes::new_with(py, deriver.len()?, |b| {
+        let len = deriver.len()?;
+        Ok(pyo3::types::PyBytes::new_with(py, len, |b| {
             let n = deriver.derive(b).map_err(|_| {
                 pyo3::exceptions::PyValueError::new_err("Error computing shared key.")
             })?;
@@ -270,9 +271,9 @@ impl ECPrivateKey {
         &self,
         py: pyo3::Python<'p>,
         data: CffiBuf<'_>,
-        signature_algorithm: &pyo3::PyAny,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
-        if !signature_algorithm.is_instance(types::ECDSA.get(py)?)? {
+        signature_algorithm: pyo3::Bound<'_, pyo3::PyAny>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        if !signature_algorithm.is_instance(&types::ECDSA.get(py)?)? {
             return Err(CryptographyError::from(
                 exceptions::UnsupportedAlgorithm::new_err((
                     "Unsupported elliptic curve signature algorithm",
@@ -280,21 +281,40 @@ impl ECPrivateKey {
                 )),
             ));
         }
-
-        let (data, _) = utils::calculate_digest_and_algorithm(
-            py,
-            data.as_bytes(),
-            signature_algorithm.getattr(pyo3::intern!(py, "algorithm"))?,
-        )?;
+        let bound_algorithm = signature_algorithm.getattr(pyo3::intern!(py, "algorithm"))?;
+        let (data, algo) =
+            utils::calculate_digest_and_algorithm(py, data.as_bytes(), &bound_algorithm)?;
 
         let mut signer = openssl::pkey_ctx::PkeyCtx::new(&self.pkey)?;
         signer.sign_init()?;
+        cfg_if::cfg_if! {
+            if #[cfg(CRYPTOGRAPHY_OPENSSL_320_OR_GREATER)]{
+                let deterministic: bool = signature_algorithm
+                    .getattr(pyo3::intern!(py, "deterministic_signing"))?
+                    .extract()?;
+                if deterministic {
+                    let hash_function_name = algo
+                        .getattr(pyo3::intern!(py, "name"))?
+                        .extract::<pyo3::pybacked::PyBackedStr>()?;
+                    let hash_function = openssl::md::Md::fetch(None, &hash_function_name, None)?;
+                    // Setting a deterministic nonce type requires to explicitly set the hash function.
+                    // See https://github.com/openssl/openssl/issues/23205
+                    signer.set_signature_md(&hash_function)?;
+                    signer.set_nonce_type(openssl::pkey_ctx::NonceType::DETERMINISTIC_K)?;
+                } else {
+                    signer.set_nonce_type(openssl::pkey_ctx::NonceType::RANDOM_K)?;
+                }
+            } else {
+                let _ = algo;
+            }
+        }
+
         // TODO: This does an extra allocation and copy. This can't easily use
         // `PyBytes::new_with` because the exact length of the signature isn't
         // easily known a priori (if `r` or `s` has a leading 0, the signature
         // will be a byte or two shorter than the maximum possible length).
         let mut sig = vec![];
-        signer.sign_to_vec(data, &mut sig)?;
+        signer.sign_to_vec(data.as_bytes(), &mut sig)?;
         Ok(pyo3::types::PyBytes::new(py, &sig))
     }
 
@@ -338,12 +358,12 @@ impl ECPrivateKey {
     }
 
     fn private_bytes<'p>(
-        slf: &pyo3::PyCell<Self>,
+        slf: &pyo3::Bound<'p, Self>,
         py: pyo3::Python<'p>,
-        encoding: &pyo3::PyAny,
-        format: &pyo3::PyAny,
-        encryption_algorithm: &pyo3::PyAny,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+        encoding: &pyo3::Bound<'p, pyo3::PyAny>,
+        format: &pyo3::Bound<'p, pyo3::PyAny>,
+        encryption_algorithm: &pyo3::Bound<'p, pyo3::PyAny>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         utils::pkey_private_bytes(
             py,
             slf,
@@ -357,11 +377,14 @@ impl ECPrivateKey {
     }
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl ECPublicKey {
     #[getter]
-    fn key_size<'p>(&'p self, py: pyo3::Python<'p>) -> pyo3::PyResult<&'p pyo3::PyAny> {
-        self.curve.as_ref(py).getattr(pyo3::intern!(py, "key_size"))
+    fn key_size<'p>(
+        &'p self,
+        py: pyo3::Python<'p>,
+    ) -> pyo3::PyResult<pyo3::Bound<'p, pyo3::PyAny>> {
+        self.curve.bind(py).getattr(pyo3::intern!(py, "key_size"))
     }
 
     fn verify(
@@ -369,9 +392,9 @@ impl ECPublicKey {
         py: pyo3::Python<'_>,
         signature: CffiBuf<'_>,
         data: CffiBuf<'_>,
-        signature_algorithm: &pyo3::PyAny,
+        signature_algorithm: pyo3::Bound<'_, pyo3::PyAny>,
     ) -> CryptographyResult<()> {
-        if !signature_algorithm.is_instance(types::ECDSA.get(py)?)? {
+        if !signature_algorithm.is_instance(&types::ECDSA.get(py)?)? {
             return Err(CryptographyError::from(
                 exceptions::UnsupportedAlgorithm::new_err((
                     "Unsupported elliptic curve signature algorithm",
@@ -383,12 +406,14 @@ impl ECPublicKey {
         let (data, _) = utils::calculate_digest_and_algorithm(
             py,
             data.as_bytes(),
-            signature_algorithm.getattr(pyo3::intern!(py, "algorithm"))?,
+            &signature_algorithm.getattr(pyo3::intern!(py, "algorithm"))?,
         )?;
 
         let mut verifier = openssl::pkey_ctx::PkeyCtx::new(&self.pkey)?;
         verifier.verify_init()?;
-        let valid = verifier.verify(data, signature.as_bytes()).unwrap_or(false);
+        let valid = verifier
+            .verify(data.as_bytes(), signature.as_bytes())
+            .unwrap_or(false);
         if !valid {
             return Err(CryptographyError::from(
                 exceptions::InvalidSignature::new_err(()),
@@ -420,11 +445,11 @@ impl ECPublicKey {
     }
 
     fn public_bytes<'p>(
-        slf: &pyo3::PyCell<Self>,
+        slf: &pyo3::Bound<'p, Self>,
         py: pyo3::Python<'p>,
-        encoding: &pyo3::PyAny,
-        format: &pyo3::PyAny,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+        encoding: &pyo3::Bound<'p, pyo3::PyAny>,
+        format: &pyo3::Bound<'p, pyo3::PyAny>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         utils::pkey_public_bytes(py, slf, &slf.borrow().pkey, encoding, format, true, false)
     }
 
@@ -437,20 +462,20 @@ impl ECPublicKey {
     }
 }
 
-#[pyo3::prelude::pyclass(frozen, module = "cryptography.hazmat.primitives.asymmetric.ec")]
+#[pyo3::pyclass(frozen, module = "cryptography.hazmat.primitives.asymmetric.ec")]
 struct EllipticCurvePrivateNumbers {
     #[pyo3(get)]
-    private_value: pyo3::Py<pyo3::types::PyLong>,
+    private_value: pyo3::Py<pyo3::types::PyInt>,
     #[pyo3(get)]
     public_numbers: pyo3::Py<EllipticCurvePublicNumbers>,
 }
 
-#[pyo3::prelude::pyclass(frozen, module = "cryptography.hazmat.primitives.asymmetric.ec")]
+#[pyo3::pyclass(frozen, module = "cryptography.hazmat.primitives.asymmetric.ec")]
 struct EllipticCurvePublicNumbers {
     #[pyo3(get)]
-    x: pyo3::Py<pyo3::types::PyLong>,
+    x: pyo3::Py<pyo3::types::PyInt>,
     #[pyo3(get)]
-    y: pyo3::Py<pyo3::types::PyLong>,
+    y: pyo3::Py<pyo3::types::PyInt>,
     #[pyo3(get)]
     curve: pyo3::Py<pyo3::PyAny>,
 }
@@ -460,8 +485,7 @@ fn public_key_from_numbers(
     numbers: &EllipticCurvePublicNumbers,
     curve: &openssl::ec::EcGroupRef,
 ) -> CryptographyResult<openssl::ec::EcKey<openssl::pkey::Public>> {
-    let zero = (0).to_object(py);
-    if numbers.x.as_ref(py).lt(&zero)? || numbers.y.as_ref(py).lt(&zero)? {
+    if numbers.x.bind(py).lt(0)? || numbers.y.bind(py).lt(0)? {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyValueError::new_err(
                 "Invalid EC key. Both x and y must be non-negative.",
@@ -469,8 +493,8 @@ fn public_key_from_numbers(
         ));
     }
 
-    let x = utils::py_int_to_bn(py, numbers.x.as_ref(py))?;
-    let y = utils::py_int_to_bn(py, numbers.y.as_ref(py))?;
+    let x = utils::py_int_to_bn(py, numbers.x.bind(py))?;
+    let y = utils::py_int_to_bn(py, numbers.y.bind(py))?;
 
     let mut point = openssl::ec::EcPoint::new(curve)?;
     let mut bn_ctx = openssl::bn::BigNumContext::new()?;
@@ -485,11 +509,11 @@ fn public_key_from_numbers(
     Ok(openssl::ec::EcKey::from_public_key(curve, &point)?)
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl EllipticCurvePrivateNumbers {
     #[new]
     fn new(
-        private_value: pyo3::Py<pyo3::types::PyLong>,
+        private_value: pyo3::Py<pyo3::types::PyInt>,
         public_numbers: pyo3::Py<EllipticCurvePublicNumbers>,
     ) -> EllipticCurvePrivateNumbers {
         EllipticCurvePrivateNumbers {
@@ -498,16 +522,18 @@ impl EllipticCurvePrivateNumbers {
         }
     }
 
+    #[pyo3(signature = (backend=None))]
     fn private_key(
         &self,
         py: pyo3::Python<'_>,
-        backend: Option<&pyo3::PyAny>,
+        backend: Option<pyo3::Bound<'_, pyo3::PyAny>>,
     ) -> CryptographyResult<ECPrivateKey> {
         let _ = backend;
 
-        let curve = curve_from_py_curve(py, self.public_numbers.get().curve.as_ref(py), false)?;
+        let curve =
+            curve_from_py_curve(py, self.public_numbers.get().curve.bind(py).clone(), false)?;
         let public_key = public_key_from_numbers(py, self.public_numbers.get(), &curve)?;
-        let private_value = utils::py_int_to_bn(py, self.private_value.as_ref(py))?;
+        let private_value = utils::py_int_to_bn(py, self.private_value.bind(py))?;
 
         let mut bn_ctx = openssl::bn::BigNumContext::new()?;
         let mut expected_pub = openssl::ec::EcPoint::new(&curve)?;
@@ -538,36 +564,35 @@ impl EllipticCurvePrivateNumbers {
         py: pyo3::Python<'_>,
         other: pyo3::PyRef<'_, Self>,
     ) -> CryptographyResult<bool> {
-        Ok(self
-            .private_value
-            .as_ref(py)
-            .eq(other.private_value.as_ref(py))?
-            && self
-                .public_numbers
-                .as_ref(py)
-                .eq(other.public_numbers.as_ref(py))?)
+        Ok(
+            (**self.private_value.bind(py)).eq(other.private_value.bind(py))?
+                && self
+                    .public_numbers
+                    .bind(py)
+                    .eq(other.public_numbers.bind(py))?,
+        )
     }
 
     fn __hash__(&self, py: pyo3::Python<'_>) -> CryptographyResult<u64> {
         let mut hasher = DefaultHasher::new();
-        self.private_value.as_ref(py).hash()?.hash(&mut hasher);
-        self.public_numbers.as_ref(py).hash()?.hash(&mut hasher);
+        self.private_value.bind(py).hash()?.hash(&mut hasher);
+        self.public_numbers.bind(py).hash()?.hash(&mut hasher);
         Ok(hasher.finish())
     }
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl EllipticCurvePublicNumbers {
     #[new]
     fn new(
         py: pyo3::Python<'_>,
-        x: pyo3::Py<pyo3::types::PyLong>,
-        y: pyo3::Py<pyo3::types::PyLong>,
+        x: pyo3::Py<pyo3::types::PyInt>,
+        y: pyo3::Py<pyo3::types::PyInt>,
         curve: pyo3::Py<pyo3::PyAny>,
     ) -> CryptographyResult<EllipticCurvePublicNumbers> {
         if !curve
-            .as_ref(py)
-            .is_instance(types::ELLIPTIC_CURVE.get(py)?)?
+            .bind(py)
+            .is_instance(&types::ELLIPTIC_CURVE.get(py)?)?
         {
             return Err(CryptographyError::from(
                 pyo3::exceptions::PyTypeError::new_err(
@@ -579,14 +604,15 @@ impl EllipticCurvePublicNumbers {
         Ok(EllipticCurvePublicNumbers { x, y, curve })
     }
 
+    #[pyo3(signature = (backend=None))]
     fn public_key(
         &self,
         py: pyo3::Python<'_>,
-        backend: Option<&pyo3::PyAny>,
+        backend: Option<pyo3::Bound<'_, pyo3::PyAny>>,
     ) -> CryptographyResult<ECPublicKey> {
         let _ = backend;
 
-        let curve = curve_from_py_curve(py, self.curve.as_ref(py), false)?;
+        let curve = curve_from_py_curve(py, self.curve.bind(py).clone(), false)?;
         let public_key = public_key_from_numbers(py, self, &curve)?;
 
         let pkey = openssl::pkey::PKey::from_ec_key(public_key)?;
@@ -602,34 +628,34 @@ impl EllipticCurvePublicNumbers {
         py: pyo3::Python<'_>,
         other: pyo3::PyRef<'_, Self>,
     ) -> CryptographyResult<bool> {
-        Ok(self.x.as_ref(py).eq(other.x.as_ref(py))?
-            && self.y.as_ref(py).eq(other.y.as_ref(py))?
+        Ok((**self.x.bind(py)).eq(other.x.bind(py))?
+            && (**self.y.bind(py)).eq(other.y.bind(py))?
             && self
                 .curve
-                .as_ref(py)
+                .bind(py)
                 .getattr(pyo3::intern!(py, "name"))?
-                .eq(other.curve.as_ref(py).getattr(pyo3::intern!(py, "name"))?)?
+                .eq(other.curve.bind(py).getattr(pyo3::intern!(py, "name"))?)?
             && self
                 .curve
-                .as_ref(py)
+                .bind(py)
                 .getattr(pyo3::intern!(py, "key_size"))?
                 .eq(other
                     .curve
-                    .as_ref(py)
+                    .bind(py)
                     .getattr(pyo3::intern!(py, "key_size"))?)?)
     }
 
     fn __hash__(&self, py: pyo3::Python<'_>) -> CryptographyResult<u64> {
         let mut hasher = DefaultHasher::new();
-        self.x.as_ref(py).hash()?.hash(&mut hasher);
-        self.y.as_ref(py).hash()?.hash(&mut hasher);
+        self.x.bind(py).hash()?.hash(&mut hasher);
+        self.y.bind(py).hash()?.hash(&mut hasher);
         self.curve
-            .as_ref(py)
+            .bind(py)
             .getattr(pyo3::intern!(py, "name"))?
             .hash()?
             .hash(&mut hasher);
         self.curve
-            .as_ref(py)
+            .bind(py)
             .getattr(pyo3::intern!(py, "key_size"))?
             .hash()?
             .hash(&mut hasher);
@@ -637,26 +663,20 @@ impl EllipticCurvePublicNumbers {
     }
 
     fn __repr__(&self, py: pyo3::Python<'_>) -> pyo3::PyResult<String> {
-        let x = self.x.as_ref(py);
-        let y = self.y.as_ref(py);
-        let curve_name = self.curve.as_ref(py).getattr(pyo3::intern!(py, "name"))?;
+        let x = self.x.bind(py);
+        let y = self.y.bind(py);
+        let curve_name = self.curve.bind(py).getattr(pyo3::intern!(py, "name"))?;
         Ok(format!(
             "<EllipticCurvePublicNumbers(curve={curve_name}, x={x}, y={y})>"
         ))
     }
 }
 
-pub(crate) fn create_module(py: pyo3::Python<'_>) -> pyo3::PyResult<&pyo3::prelude::PyModule> {
-    let m = pyo3::prelude::PyModule::new(py, "ec")?;
-    m.add_function(pyo3::wrap_pyfunction!(curve_supported, m)?)?;
-    m.add_function(pyo3::wrap_pyfunction!(generate_private_key, m)?)?;
-    m.add_function(pyo3::wrap_pyfunction!(derive_private_key, m)?)?;
-    m.add_function(pyo3::wrap_pyfunction!(from_public_bytes, m)?)?;
-
-    m.add_class::<ECPrivateKey>()?;
-    m.add_class::<ECPublicKey>()?;
-    m.add_class::<EllipticCurvePrivateNumbers>()?;
-    m.add_class::<EllipticCurvePublicNumbers>()?;
-
-    Ok(m)
+#[pyo3::pymodule]
+pub(crate) mod ec {
+    #[pymodule_export]
+    use super::{
+        curve_supported, derive_private_key, from_public_bytes, generate_private_key, ECPrivateKey,
+        ECPublicKey, EllipticCurvePrivateNumbers, EllipticCurvePublicNumbers,
+    };
 }
