@@ -2,33 +2,35 @@
 // 2.0, and the BSD License. See the LICENSE file in the root of this repository
 // for complete details.
 
+use pyo3::types::{PyAnyMethods, PyBytesMethods};
+
 use crate::backend::hashes::Hash;
 use crate::error::{CryptographyError, CryptographyResult};
 use crate::{error, types};
 
 pub(crate) fn py_int_to_bn(
     py: pyo3::Python<'_>,
-    v: &pyo3::PyAny,
+    v: &pyo3::Bound<'_, pyo3::PyAny>,
 ) -> CryptographyResult<openssl::bn::BigNum> {
     let n = v
         .call_method0(pyo3::intern!(py, "bit_length"))?
         .extract::<usize>()?
         / 8
         + 1;
-    let bytes: &[u8] = v
+    let bytes = v
         .call_method1(pyo3::intern!(py, "to_bytes"), (n, pyo3::intern!(py, "big")))?
-        .extract()?;
+        .extract::<pyo3::pybacked::PyBackedBytes>()?;
 
-    Ok(openssl::bn::BigNum::from_slice(bytes)?)
+    Ok(openssl::bn::BigNum::from_slice(&bytes)?)
 }
 
 pub(crate) fn bn_to_py_int<'p>(
     py: pyo3::Python<'p>,
     b: &openssl::bn::BigNumRef,
-) -> CryptographyResult<&'p pyo3::PyAny> {
+) -> CryptographyResult<pyo3::Bound<'p, pyo3::PyAny>> {
     assert!(!b.is_negative());
 
-    let int_type = py.get_type::<pyo3::types::PyLong>();
+    let int_type = py.get_type::<pyo3::types::PyInt>();
     Ok(int_type.call_method1(
         pyo3::intern!(py, "from_bytes"),
         (b.to_vec(), pyo3::intern!(py, "big")),
@@ -42,29 +44,29 @@ pub(crate) fn bn_to_big_endian_bytes(b: &openssl::bn::BigNumRef) -> Cryptography
 #[allow(clippy::too_many_arguments)]
 pub(crate) fn pkey_private_bytes<'p>(
     py: pyo3::Python<'p>,
-    key_obj: &pyo3::PyAny,
+    key_obj: &pyo3::Bound<'p, pyo3::PyAny>,
     pkey: &openssl::pkey::PKey<openssl::pkey::Private>,
-    encoding: &pyo3::PyAny,
-    format: &pyo3::PyAny,
-    encryption_algorithm: &pyo3::PyAny,
+    encoding: &pyo3::Bound<'p, pyo3::PyAny>,
+    format: &pyo3::Bound<'p, pyo3::PyAny>,
+    encryption_algorithm: &pyo3::Bound<'p, pyo3::PyAny>,
     openssh_allowed: bool,
     raw_allowed: bool,
-) -> CryptographyResult<&'p pyo3::types::PyBytes> {
-    if !encoding.is_instance(types::ENCODING.get(py)?)? {
+) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+    if !encoding.is_instance(&types::ENCODING.get(py)?)? {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyTypeError::new_err(
                 "encoding must be an item from the Encoding enum",
             ),
         ));
     }
-    if !format.is_instance(types::PRIVATE_FORMAT.get(py)?)? {
+    if !format.is_instance(&types::PRIVATE_FORMAT.get(py)?)? {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyTypeError::new_err(
                 "format must be an item from the PrivateFormat enum",
             ),
         ));
     }
-    if !encryption_algorithm.is_instance(types::KEY_SERIALIZATION_ENCRYPTION.get(py)?)? {
+    if !encryption_algorithm.is_instance(&types::KEY_SERIALIZATION_ENCRYPTION.get(py)?)? {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyTypeError::new_err(
                 "Encryption algorithm must be a KeySerializationEncryption instance",
@@ -73,12 +75,12 @@ pub(crate) fn pkey_private_bytes<'p>(
     }
 
     if raw_allowed
-        && (encoding.is(types::ENCODING_RAW.get(py)?)
-            || format.is(types::PRIVATE_FORMAT_RAW.get(py)?))
+        && (encoding.is(&types::ENCODING_RAW.get(py)?)
+            || format.is(&types::PRIVATE_FORMAT_RAW.get(py)?))
     {
-        if !encoding.is(types::ENCODING_RAW.get(py)?)
-            || !format.is(types::PRIVATE_FORMAT_RAW.get(py)?)
-            || !encryption_algorithm.is_instance(types::NO_ENCRYPTION.get(py)?)?
+        if !encoding.is(&types::ENCODING_RAW.get(py)?)
+            || !format.is(&types::PRIVATE_FORMAT_RAW.get(py)?)
+            || !encryption_algorithm.is_instance(&types::NO_ENCRYPTION.get(py)?)?
         {
             return Err(CryptographyError::from(pyo3::exceptions::PyValueError::new_err(
                     "When using Raw both encoding and format must be Raw and encryption_algorithm must be NoEncryption()"
@@ -88,17 +90,19 @@ pub(crate) fn pkey_private_bytes<'p>(
         return Ok(pyo3::types::PyBytes::new(py, &raw_bytes));
     }
 
-    let password = if encryption_algorithm.is_instance(types::NO_ENCRYPTION.get(py)?)? {
-        b""
-    } else if encryption_algorithm.is_instance(types::BEST_AVAILABLE_ENCRYPTION.get(py)?)?
-        || (encryption_algorithm.is_instance(types::ENCRYPTION_BUILDER.get(py)?)?
+    let py_password;
+    let password = if encryption_algorithm.is_instance(&types::NO_ENCRYPTION.get(py)?)? {
+        b"" as &[u8]
+    } else if encryption_algorithm.is_instance(&types::BEST_AVAILABLE_ENCRYPTION.get(py)?)?
+        || (encryption_algorithm.is_instance(&types::ENCRYPTION_BUILDER.get(py)?)?
             && encryption_algorithm
                 .getattr(pyo3::intern!(py, "_format"))?
                 .is(format))
     {
-        encryption_algorithm
+        py_password = encryption_algorithm
             .getattr(pyo3::intern!(py, "password"))?
-            .extract::<&[u8]>()?
+            .extract::<pyo3::pybacked::PyBackedBytes>()?;
+        &py_password
     } else {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyValueError::new_err("Unsupported encryption type"),
@@ -113,8 +117,8 @@ pub(crate) fn pkey_private_bytes<'p>(
         ));
     }
 
-    if format.is(types::PRIVATE_FORMAT_PKCS8.get(py)?) {
-        if encoding.is(types::ENCODING_PEM.get(py)?) {
+    if format.is(&types::PRIVATE_FORMAT_PKCS8.get(py)?) {
+        if encoding.is(&types::ENCODING_PEM.get(py)?) {
             let pem_bytes = if password.is_empty() {
                 pkey.private_key_to_pem_pkcs8()?
             } else {
@@ -124,7 +128,7 @@ pub(crate) fn pkey_private_bytes<'p>(
                 )?
             };
             return Ok(pyo3::types::PyBytes::new(py, &pem_bytes));
-        } else if encoding.is(types::ENCODING_DER.get(py)?) {
+        } else if encoding.is(&types::ENCODING_DER.get(py)?) {
             let der_bytes = if password.is_empty() {
                 pkey.private_key_to_pkcs8()?
             } else {
@@ -140,9 +144,16 @@ pub(crate) fn pkey_private_bytes<'p>(
         ));
     }
 
-    if format.is(types::PRIVATE_FORMAT_TRADITIONAL_OPENSSL.get(py)?) {
+    if format.is(&types::PRIVATE_FORMAT_TRADITIONAL_OPENSSL.get(py)?) {
+        if cryptography_openssl::fips::is_enabled() && !password.is_empty() {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err(
+                    "Encrypted traditional OpenSSL format is not supported in FIPS mode",
+                ),
+            ));
+        }
         if let Ok(rsa) = pkey.rsa() {
-            if encoding.is(types::ENCODING_PEM.get(py)?) {
+            if encoding.is(&types::ENCODING_PEM.get(py)?) {
                 let pem_bytes = if password.is_empty() {
                     rsa.private_key_to_pem()?
                 } else {
@@ -152,7 +163,7 @@ pub(crate) fn pkey_private_bytes<'p>(
                     )?
                 };
                 return Ok(pyo3::types::PyBytes::new(py, &pem_bytes));
-            } else if encoding.is(types::ENCODING_DER.get(py)?) {
+            } else if encoding.is(&types::ENCODING_DER.get(py)?) {
                 if !password.is_empty() {
                     return Err(CryptographyError::from(
                         pyo3::exceptions::PyValueError::new_err(
@@ -165,7 +176,7 @@ pub(crate) fn pkey_private_bytes<'p>(
                 return Ok(pyo3::types::PyBytes::new(py, &der_bytes));
             }
         } else if let Ok(dsa) = pkey.dsa() {
-            if encoding.is(types::ENCODING_PEM.get(py)?) {
+            if encoding.is(&types::ENCODING_PEM.get(py)?) {
                 let pem_bytes = if password.is_empty() {
                     dsa.private_key_to_pem()?
                 } else {
@@ -175,7 +186,7 @@ pub(crate) fn pkey_private_bytes<'p>(
                     )?
                 };
                 return Ok(pyo3::types::PyBytes::new(py, &pem_bytes));
-            } else if encoding.is(types::ENCODING_DER.get(py)?) {
+            } else if encoding.is(&types::ENCODING_DER.get(py)?) {
                 if !password.is_empty() {
                     return Err(CryptographyError::from(
                         pyo3::exceptions::PyValueError::new_err(
@@ -188,7 +199,7 @@ pub(crate) fn pkey_private_bytes<'p>(
                 return Ok(pyo3::types::PyBytes::new(py, &der_bytes));
             }
         } else if let Ok(ec) = pkey.ec_key() {
-            if encoding.is(types::ENCODING_PEM.get(py)?) {
+            if encoding.is(&types::ENCODING_PEM.get(py)?) {
                 let pem_bytes = if password.is_empty() {
                     ec.private_key_to_pem()?
                 } else {
@@ -198,7 +209,7 @@ pub(crate) fn pkey_private_bytes<'p>(
                     )?
                 };
                 return Ok(pyo3::types::PyBytes::new(py, &pem_bytes));
-            } else if encoding.is(types::ENCODING_DER.get(py)?) {
+            } else if encoding.is(&types::ENCODING_DER.get(py)?) {
                 if !password.is_empty() {
                     return Err(CryptographyError::from(
                         pyo3::exceptions::PyValueError::new_err(
@@ -214,8 +225,8 @@ pub(crate) fn pkey_private_bytes<'p>(
     }
 
     // OpenSSH + PEM
-    if openssh_allowed && format.is(types::PRIVATE_FORMAT_OPENSSH.get(py)?) {
-        if encoding.is(types::ENCODING_PEM.get(py)?) {
+    if openssh_allowed && format.is(&types::PRIVATE_FORMAT_OPENSSH.get(py)?) {
+        if encoding.is(&types::ENCODING_PEM.get(py)?) {
             return Ok(types::SERIALIZE_SSH_PRIVATE_KEY
                 .get(py)?
                 .call1((key_obj, password, encryption_algorithm))?
@@ -236,21 +247,21 @@ pub(crate) fn pkey_private_bytes<'p>(
 
 pub(crate) fn pkey_public_bytes<'p>(
     py: pyo3::Python<'p>,
-    key_obj: &pyo3::PyAny,
+    key_obj: &pyo3::Bound<'p, pyo3::PyAny>,
     pkey: &openssl::pkey::PKey<openssl::pkey::Public>,
-    encoding: &pyo3::PyAny,
-    format: &pyo3::PyAny,
+    encoding: &pyo3::Bound<'p, pyo3::PyAny>,
+    format: &pyo3::Bound<'p, pyo3::PyAny>,
     openssh_allowed: bool,
     raw_allowed: bool,
-) -> CryptographyResult<&'p pyo3::types::PyBytes> {
-    if !encoding.is_instance(types::ENCODING.get(py)?)? {
+) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+    if !encoding.is_instance(&types::ENCODING.get(py)?)? {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyTypeError::new_err(
                 "encoding must be an item from the Encoding enum",
             ),
         ));
     }
-    if !format.is_instance(types::PUBLIC_FORMAT.get(py)?)? {
+    if !format.is_instance(&types::PUBLIC_FORMAT.get(py)?)? {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyTypeError::new_err(
                 "format must be an item from the PublicFormat enum",
@@ -259,11 +270,11 @@ pub(crate) fn pkey_public_bytes<'p>(
     }
 
     if raw_allowed
-        && (encoding.is(types::ENCODING_RAW.get(py)?)
-            || format.is(types::PUBLIC_FORMAT_RAW.get(py)?))
+        && (encoding.is(&types::ENCODING_RAW.get(py)?)
+            || format.is(&types::PUBLIC_FORMAT_RAW.get(py)?))
     {
-        if !encoding.is(types::ENCODING_RAW.get(py)?)
-            || !format.is(types::PUBLIC_FORMAT_RAW.get(py)?)
+        if !encoding.is(&types::ENCODING_RAW.get(py)?)
+            || !format.is(&types::PUBLIC_FORMAT_RAW.get(py)?)
         {
             return Err(CryptographyError::from(
                 pyo3::exceptions::PyValueError::new_err(
@@ -276,11 +287,11 @@ pub(crate) fn pkey_public_bytes<'p>(
     }
 
     // SubjectPublicKeyInfo + PEM/DER
-    if format.is(types::PUBLIC_FORMAT_SUBJECT_PUBLIC_KEY_INFO.get(py)?) {
-        if encoding.is(types::ENCODING_PEM.get(py)?) {
+    if format.is(&types::PUBLIC_FORMAT_SUBJECT_PUBLIC_KEY_INFO.get(py)?) {
+        if encoding.is(&types::ENCODING_PEM.get(py)?) {
             let pem_bytes = pkey.public_key_to_pem()?;
             return Ok(pyo3::types::PyBytes::new(py, &pem_bytes));
-        } else if encoding.is(types::ENCODING_DER.get(py)?) {
+        } else if encoding.is(&types::ENCODING_DER.get(py)?) {
             let der_bytes = pkey.public_key_to_der()?;
             return Ok(pyo3::types::PyBytes::new(py, &der_bytes));
         }
@@ -292,10 +303,10 @@ pub(crate) fn pkey_public_bytes<'p>(
     }
 
     if let Ok(ec) = pkey.ec_key() {
-        if encoding.is(types::ENCODING_X962.get(py)?) {
-            let point_form = if format.is(types::PUBLIC_FORMAT_UNCOMPRESSED_POINT.get(py)?) {
+        if encoding.is(&types::ENCODING_X962.get(py)?) {
+            let point_form = if format.is(&types::PUBLIC_FORMAT_UNCOMPRESSED_POINT.get(py)?) {
                 openssl::ec::PointConversionForm::UNCOMPRESSED
-            } else if format.is(types::PUBLIC_FORMAT_COMPRESSED_POINT.get(py)?) {
+            } else if format.is(&types::PUBLIC_FORMAT_COMPRESSED_POINT.get(py)?) {
                 openssl::ec::PointConversionForm::COMPRESSED
             } else {
                 return Err(CryptographyError::from(
@@ -313,11 +324,11 @@ pub(crate) fn pkey_public_bytes<'p>(
     }
 
     if let Ok(rsa) = pkey.rsa() {
-        if format.is(types::PUBLIC_FORMAT_PKCS1.get(py)?) {
-            if encoding.is(types::ENCODING_PEM.get(py)?) {
+        if format.is(&types::PUBLIC_FORMAT_PKCS1.get(py)?) {
+            if encoding.is(&types::ENCODING_PEM.get(py)?) {
                 let pem_bytes = rsa.public_key_to_pem_pkcs1()?;
                 return Ok(pyo3::types::PyBytes::new(py, &pem_bytes));
-            } else if encoding.is(types::ENCODING_DER.get(py)?) {
+            } else if encoding.is(&types::ENCODING_DER.get(py)?) {
                 let der_bytes = rsa.public_key_to_der_pkcs1()?;
                 return Ok(pyo3::types::PyBytes::new(py, &der_bytes));
             }
@@ -330,8 +341,8 @@ pub(crate) fn pkey_public_bytes<'p>(
     }
 
     // OpenSSH + OpenSSH
-    if openssh_allowed && format.is(types::PUBLIC_FORMAT_OPENSSH.get(py)?) {
-        if encoding.is(types::ENCODING_OPENSSH.get(py)?) {
+    if openssh_allowed && format.is(&types::PUBLIC_FORMAT_OPENSSH.get(py)?) {
+        if encoding.is(&types::ENCODING_OPENSSH.get(py)?) {
             return Ok(types::SERIALIZE_SSH_PUBLIC_KEY
                 .get(py)?
                 .call1((key_obj,))?
@@ -350,22 +361,39 @@ pub(crate) fn pkey_public_bytes<'p>(
     ))
 }
 
+pub(crate) enum BytesOrPyBytes<'a> {
+    Bytes(&'a [u8]),
+    PyBytes(pyo3::Bound<'a, pyo3::types::PyBytes>),
+}
+
+impl BytesOrPyBytes<'_> {
+    pub(crate) fn as_bytes(&self) -> &[u8] {
+        match self {
+            BytesOrPyBytes::Bytes(v) => v,
+            BytesOrPyBytes::PyBytes(v) => v.as_bytes(),
+        }
+    }
+}
+
 pub(crate) fn calculate_digest_and_algorithm<'p>(
     py: pyo3::Python<'p>,
-    mut data: &'p [u8],
-    mut algorithm: &'p pyo3::PyAny,
-) -> CryptographyResult<(&'p [u8], &'p pyo3::PyAny)> {
-    if algorithm.is_instance(types::PREHASHED.get(py)?)? {
-        algorithm = algorithm.getattr("_algorithm")?;
+    data: &'p [u8],
+    algorithm: &pyo3::Bound<'p, pyo3::PyAny>,
+) -> CryptographyResult<(BytesOrPyBytes<'p>, pyo3::Bound<'p, pyo3::PyAny>)> {
+    let (algorithm, data) = if algorithm.is_instance(&types::PREHASHED.get(py)?)? {
+        (
+            algorithm.getattr("_algorithm")?,
+            BytesOrPyBytes::Bytes(data),
+        )
     } else {
         // Potential optimization: rather than allocate a PyBytes in
         // `h.finalize()`, have a way to get the `DigestBytes` directly.
         let mut h = Hash::new(py, algorithm, None)?;
         h.update_bytes(data)?;
-        data = h.finalize(py)?.as_bytes();
-    }
+        (algorithm.clone(), BytesOrPyBytes::PyBytes(h.finalize(py)?))
+    };
 
-    if data.len() != algorithm.getattr("digest_size")?.extract()? {
+    if data.as_bytes().len() != (algorithm.getattr("digest_size")?.extract::<usize>()?) {
         return Err(CryptographyError::from(
             pyo3::exceptions::PyValueError::new_err(
                 "The provided data must be the same length as the hash algorithm's digest size.",
@@ -429,12 +457,12 @@ pub(crate) fn handle_key_load_result<T>(
             )),
         )),
         (Err(e), _, _) => {
-            let errors = error::list_from_openssl_error(py, e);
+            let errors = error::list_from_openssl_error(py, &e);
             Err(CryptographyError::from(
-                types::BACKEND_HANDLE_KEY_LOADING_ERROR
-                    .get(py)?
-                    .call1((errors,))
-                    .unwrap_err(),
+                pyo3::exceptions::PyValueError::new_err((
+                    "Could not deserialize key data. The data may be in an incorrect format, the provided password may be incorrect, it may be encrypted with an unsupported algorithm, or it may be an unsupported key type (e.g. EC curves with explicit parameters).",
+                    errors.unbind(),
+                ))
             ))
         }
     }
