@@ -13,8 +13,13 @@ from binascii import hexlify
 import pytest
 
 from cryptography import exceptions, utils, x509
+from cryptography.hazmat.bindings._rust import openssl as rust_openssl
 from cryptography.hazmat.primitives import hashes, serialization
 from cryptography.hazmat.primitives.asymmetric import ec
+from cryptography.hazmat.primitives.asymmetric.ec import (
+    EllipticCurvePrivateKey,
+    EllipticCurvePublicKey,
+)
 from cryptography.hazmat.primitives.asymmetric.utils import (
     Prehashed,
     encode_dss_signature,
@@ -26,6 +31,7 @@ from ...utils import (
     load_fips_ecdsa_signing_vectors,
     load_kasvs_ecdh_vectors,
     load_nist_vectors,
+    load_rfc6979_vectors,
     load_vectors_from_file,
     raises_unsupported_algorithm,
 )
@@ -46,9 +52,8 @@ def _skip_ecdsa_vector(backend, curve: ec.EllipticCurve, hash_type):
         ec.ECDSA(hash_type()), curve
     ):
         pytest.skip(
-            "ECDSA not supported with this hash {} and curve {}.".format(
-                hash_type().name, curve.name
-            )
+            f"ECDSA not supported with this hash {hash_type().name} and "
+            f"curve {curve.name}."
         )
 
 
@@ -133,12 +138,22 @@ def test_derive_point_at_infinity(backend):
     # BoringSSL rejects infinity points before it ever gets to us, so it
     # uses a more generic error message.
     match = (
-        "infinity" if not backend._lib.CRYPTOGRAPHY_IS_BORINGSSL else "Invalid"
+        "infinity" if not rust_openssl.CRYPTOGRAPHY_IS_BORINGSSL else "Invalid"
     )
     with pytest.raises(ValueError, match=match):
         ec.derive_private_key(q, ec.SECP256R1())
 
 
+def test_derive_point_invalid_key(backend):
+    curve = ec.SECP256R1()
+    _skip_curve_unsupported(backend, curve)
+    with pytest.raises(ValueError):
+        ec.derive_private_key(
+            0xE2563328DFABF68188606B91324281C1D58A4456431B09D510B35FECC9F307CA1822846FA2671371A9A81BAC0E35749D,
+            curve,
+        )
+
+
 def test_ec_numbers():
     numbers = ec.EllipticCurvePrivateNumbers(
         1, ec.EllipticCurvePublicNumbers(2, 3, DummyCurve())
@@ -423,7 +438,7 @@ class TestECDSAVectors:
         # uses a more generic error message.
         match = (
             r"infinity|invalid form"
-            if not backend._lib.CRYPTOGRAPHY_IS_BORINGSSL
+            if not rust_openssl.CRYPTOGRAPHY_IS_BORINGSSL
             else None
         )
         with pytest.raises(ValueError, match=match):
@@ -507,10 +522,92 @@ class TestECDSAVectors:
                         signature, vector["message"], ec.ECDSA(hash_type())
                     )
 
+    def test_unsupported_deterministic_nonce(self, backend):
+        if backend.ecdsa_deterministic_supported():
+            pytest.skip(
+                f"ECDSA deterministic signing is supported by this"
+                f" backend {backend}"
+            )
+        with pytest.raises(exceptions.UnsupportedAlgorithm):
+            ec.ECDSA(hashes.SHA256(), deterministic_signing=True)
+
+    def test_deterministic_nonce(self, backend, subtests):
+        if not backend.ecdsa_deterministic_supported():
+            pytest.skip(
+                f"ECDSA deterministic signing is not supported by this"
+                f" backend {backend}"
+            )
+
+        supported_hash_algorithms = {
+            "SHA1": hashes.SHA1(),
+            "SHA224": hashes.SHA224(),
+            "SHA256": hashes.SHA256(),
+            "SHA384": hashes.SHA384(),
+            "SHA512": hashes.SHA512(),
+        }
+        curves = {
+            "B-163": ec.SECT163R2(),
+            "B-233": ec.SECT233R1(),
+            "B-283": ec.SECT283R1(),
+            "B-409": ec.SECT409R1(),
+            "B-571": ec.SECT571R1(),
+            "K-163": ec.SECT163K1(),
+            "K-233": ec.SECT233K1(),
+            "K-283": ec.SECT283K1(),
+            "K-409": ec.SECT409K1(),
+            "K-571": ec.SECT571K1(),
+            "P-192": ec.SECP192R1(),
+            "P-224": ec.SECP224R1(),
+            "P-256": ec.SECP256R1(),
+            "P-384": ec.SECP384R1(),
+            "P-521": ec.SECP521R1(),
+        }
+        vectors = load_vectors_from_file(
+            os.path.join(
+                "asymmetric", "ECDSA", "RFC6979", "evppkey_ecdsa_rfc6979.txt"
+            ),
+            load_rfc6979_vectors,
+        )
+
+        for vector in vectors:
+            with subtests.test():
+                input = bytes(vector["input"], "utf-8")
+                output = bytes.fromhex(vector["output"])
+                key = bytes("\n".join(vector["key"]), "utf-8")
+                curve = curves[vector["key_name"].split("_")[0]]
+                _skip_curve_unsupported(backend, curve)
+
+                if "digest_sign" in vector:
+                    algorithm = vector["digest_sign"]
+                    hash_algorithm = supported_hash_algorithms[algorithm]
+                    algorithm = ec.ECDSA(
+                        hash_algorithm,
+                        deterministic_signing=vector["deterministic_nonce"],
+                    )
+                    private_key = serialization.load_pem_private_key(
+                        key, password=None
+                    )
+                    assert isinstance(private_key, EllipticCurvePrivateKey)
+                    signature = private_key.sign(input, algorithm)
+                    assert signature == output
+                else:
+                    assert "digest_verify" in vector
+                    algorithm = vector["digest_verify"]
+                    assert algorithm in supported_hash_algorithms
+                    hash_algorithm = supported_hash_algorithms[algorithm]
+                    algorithm = ec.ECDSA(hash_algorithm)
+                    public_key = serialization.load_pem_public_key(key)
+                    assert isinstance(public_key, EllipticCurvePublicKey)
+                    if vector["verify_error"]:
+                        with pytest.raises(exceptions.InvalidSignature):
+                            public_key.verify(output, input, algorithm)
+                    else:
+                        public_key.verify(output, input, algorithm)
+
     def test_sign(self, backend):
         _skip_curve_unsupported(backend, ec.SECP256R1())
         message = b"one little message"
-        algorithm = ec.ECDSA(hashes.SHA1())
+        algorithm = ec.ECDSA(hashes.SHA256())
         private_key = ec.generate_private_key(ec.SECP256R1(), backend)
         signature = private_key.sign(message, algorithm)
         public_key = private_key.public_key()
@@ -519,7 +616,7 @@ class TestECDSAVectors:
     def test_sign_verify_buffers(self, backend):
         _skip_curve_unsupported(backend, ec.SECP256R1())
         message = bytearray(b"one little message")
-        algorithm = ec.ECDSA(hashes.SHA1())
+        algorithm = ec.ECDSA(hashes.SHA256())
         private_key = ec.generate_private_key(ec.SECP256R1(), backend)
         signature = private_key.sign(message, algorithm)
         public_key = private_key.public_key()
@@ -528,19 +625,19 @@ class TestECDSAVectors:
     def test_sign_prehashed(self, backend):
         _skip_curve_unsupported(backend, ec.SECP256R1())
         message = b"one little message"
-        h = hashes.Hash(hashes.SHA1(), backend)
+        h = hashes.Hash(hashes.SHA256(), backend)
         h.update(message)
         data = h.finalize()
-        algorithm = ec.ECDSA(Prehashed(hashes.SHA1()))
+        algorithm = ec.ECDSA(Prehashed(hashes.SHA256()))
         private_key = ec.generate_private_key(ec.SECP256R1(), backend)
         signature = private_key.sign(data, algorithm)
         public_key = private_key.public_key()
-        public_key.verify(signature, message, ec.ECDSA(hashes.SHA1()))
+        public_key.verify(signature, message, ec.ECDSA(hashes.SHA256()))
 
     def test_sign_prehashed_digest_mismatch(self, backend):
         _skip_curve_unsupported(backend, ec.SECP256R1())
         message = b"one little message"
-        h = hashes.Hash(hashes.SHA1(), backend)
+        h = hashes.Hash(hashes.SHA224(), backend)
         h.update(message)
         data = h.finalize()
         algorithm = ec.ECDSA(Prehashed(hashes.SHA256()))
@@ -551,7 +648,7 @@ class TestECDSAVectors:
     def test_verify(self, backend):
         _skip_curve_unsupported(backend, ec.SECP256R1())
         message = b"one little message"
-        algorithm = ec.ECDSA(hashes.SHA1())
+        algorithm = ec.ECDSA(hashes.SHA256())
         private_key = ec.generate_private_key(ec.SECP256R1(), backend)
         signature = private_key.sign(message, algorithm)
         public_key = private_key.public_key()
@@ -560,20 +657,22 @@ class TestECDSAVectors:
     def test_verify_prehashed(self, backend):
         _skip_curve_unsupported(backend, ec.SECP256R1())
         message = b"one little message"
-        algorithm = ec.ECDSA(hashes.SHA1())
+        algorithm = ec.ECDSA(hashes.SHA256())
         private_key = ec.generate_private_key(ec.SECP256R1(), backend)
         signature = private_key.sign(message, algorithm)
-        h = hashes.Hash(hashes.SHA1(), backend)
+        h = hashes.Hash(hashes.SHA256(), backend)
         h.update(message)
         data = h.finalize()
         public_key = private_key.public_key()
-        public_key.verify(signature, data, ec.ECDSA(Prehashed(hashes.SHA1())))
+        public_key.verify(
+            signature, data, ec.ECDSA(Prehashed(hashes.SHA256()))
+        )
 
     def test_verify_prehashed_digest_mismatch(self, backend):
         _skip_curve_unsupported(backend, ec.SECP256R1())
         message = b"one little message"
         private_key = ec.generate_private_key(ec.SECP256R1(), backend)
-        h = hashes.Hash(hashes.SHA1(), backend)
+        h = hashes.Hash(hashes.SHA224(), backend)
         h.update(message)
         data = h.finalize()
         public_key = private_key.public_key()
@@ -684,6 +783,24 @@ class TestECSerialization:
         priv_num = key.private_numbers()
         assert loaded_priv_num == priv_num
 
+    @pytest.mark.supported(
+        only_if=lambda backend: backend._fips_enabled,
+        skip_message="Requires FIPS",
+    )
+    def test_traditional_serialization_fips(self, backend):
+        key_bytes = load_vectors_from_file(
+            os.path.join("asymmetric", "PKCS8", "ec_private_key.pem"),
+            lambda pemfile: pemfile.read().encode(),
+        )
+        key = serialization.load_pem_private_key(key_bytes, None, backend)
+        assert isinstance(key, ec.EllipticCurvePrivateKey)
+        with pytest.raises(ValueError):
+            key.private_bytes(
+                serialization.Encoding.PEM,
+                serialization.PrivateFormat.TraditionalOpenSSL,
+                serialization.BestAvailableEncryption(b"password"),
+            )
+
     @pytest.mark.parametrize(
         ("encoding", "fmt"),
         [
